from typing import Callable, List, Optional, Any
from io import BytesIO


class BinaryWriter:
    def __init__(self):
        self._bytes = bytearray()

    def write_int(self, value: int) -> None:
        self._bytes.extend(value.to_bytes(4, "little", signed=True))

    def write_long(self, value: int) -> None:
        self._bytes.extend(value.to_bytes(8, "little", signed=True))

    def write_string(self, value: str) -> None:
        encoded = value.encode("utf-8")
        length = len(encoded)
        if length <= 253:
            self._bytes.append(length)
        else:
            self._bytes.append(254)
            self._bytes.extend(length.to_bytes(3, "little"))
        self._bytes.extend(encoded)
        while len(self._bytes) % 4 != 0:
            self._bytes.append(0)

    def write_bytes(self, value: bytes) -> None:
        self._bytes.extend(value)

    def get_bytes(self) -> bytes:
        return bytes(self._bytes)


from typing import Callable, List, Optional, Any
from io import BytesIO
# from typing import TypeVar

class BinaryReader:
    def __init__(self, data: bytes):
        self.data = data
        self.pos = 0

    def read_int(self) -> int:
        result = int.from_bytes(self.data[self.pos:self.pos + 4], "little", signed=True)
        self.pos += 4
        return result

    def read_long(self) -> int:
        result = int.from_bytes(self.data[self.pos:self.pos + 8], "little", signed=True)
        self.pos += 8
        return result

    def read_string(self) -> str:
        length = self.data[self.pos]
        self.pos += 1
        if length == 254:
            length = int.from_bytes(self.data[self.pos:self.pos + 3], "little")
            self.pos += 3
        result = self.data[self.pos:self.pos + length].decode("utf-8")
        self.pos += length
        while self.pos % 4 != 0:
            self.pos += 1
        return result

    def read_bytes(self, length: int) -> bytes:
        result = self.data[self.pos:self.pos + length]
        self.pos += length
        return result


from typing import Dict, Any, Optional
from neongram.parser.tl_object import TLFunction


class AuthSendCode(TLFunction):
    def __init__(self):
        super().__init__(
            "auth.sendCode",
            0xA677244F,
            [
                {"name": "phone_number", "type": "string"},
                {"name": "api_id", "type": "int"},
                {"name": "api_hash", "type": "string"},
                {"name": "settings", "type": "CodeSettings"},
            ],
            "auth.SentCode"
        )

class AuthSignIn(TLFunction):
    def __init__(self):
        super().__init__(
            "auth.signIn",
            0x8D52A951,
            [
                {"name": "phone_number", "type": "string"},
                {"name": "phone_code_hash", "type": "string"},
                {"name": "phone_code", "type": "string"},
                {"name": "password", "type": "string", "optional": True},
            ],
            "auth.Authorization"
        )

class AuthImportBotAuthorization(TLFunction):
    def __init__(self):
        super().__init__(
            "auth.importBotAuthorization",
            0x67A3FF2C,
            [
                {"name": "api_id", "type": "int"},
                {"name": "api_hash", "type": "string"},
                {"name": "bot_auth_token", "type": "string"},
            ],
            "auth.Authorization"
        )

class UsersGetUsers(TLFunction):
    def __init__(self):
        super().__init__(
            "users.getUsers",
            0x0D91A548,
            [
                {"name": "id", "type": "Vector<InputUser>"},
            ],
            "Vector<User>"
        )


class MessagesSendMessage(TLFunction):
    def __init__(self):
        super().__init__("messages.sendMessage", 0x6f5554bb, [
            {"name": "peer", "type": "InputPeer"},
            {"name": "message", "type": "string"},
            {"name": "random_id", "type": "long"},
            {"name": "reply_to_msg_id", "type": "int", "optional": True}
        ], "Updates")


class UpdatesGetState(TLFunction):
    def __init__(self):
        super().__init__("updates.getState", 0xedd4882a, [], "updates.State")


class MessagesGetHistory(TLFunction):
    def __init__(self):
        super().__init__("messages.getHistory", 0xafa92846, [
            {"name": "peer", "type": "InputPeer"},
            {"name": "offset_id", "type": "int"},
            {"name": "offset_date", "type": "int"},
            {"name": "add_offset", "type": "int"},
            {"name": "limit", "type": "int"},
            {"name": "max_id", "type": "int", "optional": True},
            {"name": "min_id", "type": "int", "optional": True}
        ], "messages.Messages")


class AccountGetPassword(TLFunction):
    def __init__(self):
        super().__init__("account.getPassword", 0x548a30f5, [], "account.Password")


class AccountUpdatePasswordSettings(TLFunction):
    def __init__(self):
        super().__init__("account.updatePasswordSettings", 0xb60a24ed, [
            {"name": "password", "type": "InputCheckPasswordSRP"},
            {"name": "new_settings", "type": "account.PasswordInputSettings"}
        ], "Bool")


class ChannelsJoinChannel(TLFunction):
    def __init__(self):
        super().__init__("channels.joinChannel", 0x24b524c5, [
            {"name": "channel", "type": "InputChannel"}
        ], "Updates")


from typing import Optional
import socket
import ssl
import asyncio
from neongram.utils.binary_writer import BinaryWriter

class Connection:
    """Handles network connections for MTProto communication.

    Args:
        host (str): The host address.
        port (int): The port number.
        use_ssl (bool): Whether to use SSL/TLS.
    """

    def __init__(self, host: str, port: int, use_ssl: bool = True):
        self.host = host
        self.port = port
        self.use_ssl = use_ssl
        self.socket: Optional[socket.socket] = None
        self.ssl_context = self._create_ssl_context() if use_ssl else None
        self.auth_key: Optional[bytes] = None
        self.server_salt: Optional[int] = None
        self.connected = False

    def _create_ssl_context(self):
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.minimum_version = ssl.TLSVersion.TLSv1_2
        context.maximum_version = ssl.TLSVersion.TLSv1_3
        context.load_default_certs(purpose=ssl.Purpose.SERVER_AUTH)
        print(f"TLS Minimum Version: {context.minimum_version}")
        print(f"TLS Maximum Version: {context.maximum_version}")
        print(f"Supported Ciphers: {context.get_ciphers()}")
        return context

    async def connect(self) -> None:
        loop = asyncio.get_event_loop()
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            if (self.use_ssl and self.ssl_context):
                print(f"Connecting to {self.host}:{self.port} with SSL")
                self.socket = self.ssl_context.wrap_socket(
                    self.socket,
                    server_hostname=self.host,
                    do_handshake_on_connect=True  # Ensure handshake happens here
                )
                # Print SSL details after handshake
                print(f"Negotiated Protocol: {self.socket.version()}")
                print(f"Negotiated Cipher: {self.socket.cipher()}")
            await loop.run_in_executor(None, self.socket.connect, (self.host, self.port))
            self.connected = True
            print(f"Connected to {self.host}:{self.port}")
        except Exception as e:
            print(f"Connection failed: {e}")
            raise ConnectionError(f"Failed to connect: {e}")

    def send(self, data: bytes) -> None:
        if not self.socket or not self.connected:
            raise ConnectionError("No active connection")
        writer = BinaryWriter()
        writer.write_int(0xeeeeeeee)  # MTProto header
        writer.write_int(len(data))   # Length of payload
        writer.write_bytes(data)      # Payload
        self.socket.send(writer.get_bytes())

    async def receive(self, buffer_size: int = 4096) -> bytes:
        if not self.socket or not self.connected:
            raise ConnectionError("No active connection")
        loop = asyncio.get_event_loop()
        data = await loop.run_in_executor(None, self.socket.recv, buffer_size)
        if not data:
            raise ConnectionError("Connection closed by server")
        return data

    async def close(self) -> None:
        if self.socket and self.connected:
            self.socket.close()
            self.connected = False
            self.socket = None
            print("Connection closed")